# /*
# This file is part 1/2 of the Pikachu secure discussions over irc
# implementation for ircII. The 2/2 part is the file: pikapi.irc
#
# Copyright (C) 1992,99 Bisqwit (http://iki.fi/bisqwit/)
#
# This plugin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# GNU General Public License is available
# at http://www.gnu.org/copyleft/gpl.html
#
#
# This program is GNU C++ and needs to be compiled.
# It's automatically invoked by pikapi.irc, the second
# part of this implementation.
# */

#include <iostream>
#include <cstring>
#include <cctype>
#include <string>
#include <vector>
#include <map>
using namespace std;
static const char rivitunniste[] = "";
class Syntax
{
    class Choice
    {
    public:
        string text;
        unsigned target;
    };
    typedef vector<Choice> ChoiceSet;
    typedef map<unsigned, ChoiceSet> Labels;
    Labels Choices;
    void Load(const char *s)
    {
        while(*s)
        {
            unsigned label, target;
            ChoiceSet setti;
            for(label=0; isdigit(*s); ++s)label=label*10+(*s-'0');
            for(;;)
            {
                string text;
                while(*s==' '||*s=='\t')++s;
                for(; *s!='\t'&&*s!=' '; ++s)text+=*s=='_'?' ':*s;
                while(*s==' '||*s=='\t')++s;
                for(target=0; isdigit(*s); ++s)target=target*10+(*s-'0');
                if(*s != '\n')
                {
                    cerr << "No '\\n'\n" << flush;
                    throw;
                }
                ++s;
                Choice tmp = {text, target};
                
                ChoiceSet::const_iterator i;
                for(i=setti.begin(); i!=setti.end(); ++i)
                {
                	if(i->text == tmp.text.substr(0, i->text.size()))
                	{
                		cerr << "Ambiguous: \"" << i->text << "\" and \"" << tmp.text << "\"\n" << flush;
                		throw;
                	}
                }
                setti.push_back(tmp);
                if(*s!=' ' && *s!='\t')break;
            }
            
            target = setti.size();
            if((target&(target-1)) != 0)
            {
                cerr << "Found an entry with non-2^ -count of choices\n" << flush;
                throw;
            }
            Choices[label] = setti;
        }
    }
public:
    Syntax(const char *const s){Load(s);}
    bool operator< (const string &s) const
    {
        unsigned n = strlen(rivitunniste);
        return s.substr(0,n)==rivitunniste;
    }
    int operator() (const string &s) const // pakkaus
    {
        class bitti
        {
            int pos, sz;
            const unsigned char *data;
        public:
            bitti(const string &p) : pos(0), data((const unsigned char *)p.c_str()) { sz=p.size()*8; }
            inline int left() const { return sz-pos; }
            unsigned fetch(unsigned n)
            {
                unsigned res;
                unsigned bpos=0;
                for(res=0; n>0; ++pos,--n)
                    res |= ((data[pos>>3] >> (pos&7)) & 1) << bpos++;
                return res;
            }
        } data(s);
        unsigned label=0;
        cout << rivitunniste;
        while(data.left() > 0)
        {
            Labels::const_iterator i;
            i = Choices.find(label);
            if(i == Choices.end())
            {
                cerr << "Label " << label << " not found\n" << flush;
                throw;
            }
            unsigned b=0, c = i->second.size();
            while(c > (1 << b))++b;
            unsigned n = data.fetch(b);
            //  cerr << "Selection count = " << c << ", fetching " << b << " bits: " << n << endl << flush;
            cout << i->second[n].text;
            label = i->second[n].target;
        }
        cout << endl;
        return 0;
    }
    int operator- (const string &s) const // purku
    {
        unsigned label=0;
        unsigned srcpos = strlen(rivitunniste);
        class bitti
        {
            unsigned pos;
            unsigned char data;
        public:
            bitti() : pos(0), data(0) { }
            void add(unsigned res, unsigned n)
            {
                unsigned bpos=0;
                for(; n>0; --n)
                {
                    data |= ((res >> bpos++) & 1) << pos;
                    if(++pos == 8)
                    {
                        cout << data;
                        data = 0;
                        pos = 0;
                    }
                }
            }
            void finish() const
            {
                if(pos & 7)cout << data;
                cout << endl;
            }
        } data;
        
        while(srcpos < s.size())
        {
            Labels::const_iterator i;
            i = Choices.find(label);
            if(i == Choices.end())
            {
                cerr << "Label " << label << " not found\n" << flush;
                throw;
            }
            unsigned a, b=0, c=i->second.size();
            while(c > (1 << b))++b;
            for(a=0; a<c; ++a)
            {
                const string &txt = i->second[a].text;
                unsigned txtsize = txt.size();
                if(s.substr(srcpos, txtsize) == txt)
                {
                //  cerr << "Max count = " << c << ", adding " << b << " bits: " << a << endl << flush;
                    srcpos += txtsize;
                    label = i->second[a].target;
                    break;
                }
            }
            if(a == c)
            {
            	cerr << "Not found, at '" << s.substr(srcpos) << "'\n" << flush;
            	throw;            	
            }
            data.add(a, b);
        }
        data.finish();
        return 0;
    }
};

// VERSION: 1.02

int main(int argc, const char *const *argv){
Syntax is("\
0	P	1\n\
	C	7\n\
1	i	3\n\
		1\n\
3	i	4\n\
	k	5\n\
	-	13\n\
		3\n\
4	p	1\n\
	!	11\n\
	_	2\n\
	,	14\n\
5	a	6\n\
		5\n\
6	p	1\n\
	c	7\n\
		6\n\
	.	11\n\
	.	11\n\
	,	14\n\
	?	11\n\
	-	12\n\
12	p	1\n\
	c	7\n\
		12\n\
		12\n\
13	k	5\n\
		13\n\
7	h	8\n\
8	u	9\n\
9	u	10\n\
	_;)	11\n\
	...	11\n\
	.	11\n\
	_	2\n\
		10\n\
	!	11\n\
		9\n\
10	...	11\n\
	.	11\n\
	.	11\n\
	.	11\n\
	h	15\n\
	h	15\n\
	,	14\n\
	,	14\n\
15	...	11\n\
	... 14\n\
	_:)	16\n\
	_:(	16\n\
11	_:)_ 0\n\
	_	0\n\
	_	0\n\
	_	0\n\
14	_	2\n\
16	_	0\n\
	_	0\n\
2	p	1\n\
	c	7\n\
		2\n\
		2\n\
");
    string s;
    if(argc==2)s=argv[1];else getline(cin,s);
    return is<s?is-s:is(s);
}
